from scipy.interpolate import interp1d

from exorad.log import Logger
from exorad.models.signal import Radiance


class SkyFilter(Radiance):
    """
    Handler for sky filter

    Attributes
    ----------
    transmission: array
        contains the transmission information sampled at the radiance wl grid
    """
    transmission = None


class SkyForeground(Logger):

    def __init__(self, wl, description):
        super().__init__()
        self.wl = wl
        self.name = description['value']
        self.description = description
        self.skyFilter = self._get_radiance()

    def _get_radiance(self):
        skyFilter_data = self.description['data']

        em_func = interp1d(skyFilter_data['Wavelength'].to(self.wl.unit),
                           skyFilter_data['Radiance'],
                           assume_sorted=False,
                           fill_value=0.0,
                           bounds_error=False)
        skyFilter = SkyFilter(self.wl, em_func(self.wl) * skyFilter_data['Radiance'].unit)
        self.debug('radiance : {}'.format(skyFilter.data))

        tr_func = interp1d(skyFilter_data['Wavelength'].to(self.wl.unit),
                           skyFilter_data['Transmission'],
                           assume_sorted=False,
                           fill_value=1.0,
                           bounds_error=False)
        skyFilter.transmission = tr_func(self.wl)
        self.debug('transmission : {}'.format(skyFilter.transmission))

        return skyFilter



import glob
import os
import warnings

import numpy as np
from astropy import units as u
from astropy.io import ascii
from astropy.io import fits

from exorad.log import Logger
from exorad.utils import exolib
from .signal import Sed

warnings.filterwarnings('ignore', category=UserWarning, append=True)


class Star(Logger, object):
    """
      Instantiate a Stellar class using Phenix Stellar Models

      Attributes
      ----------
      lumiosity : 		float
                  Stellar bolometric luminosity computed from Phenix stellar modle. Units [W]
      wl				array
                  wavelength [micron]
      sed 			array
                  Spectral energy density [W m**-2 micron**-1]
      ph_wl			array
                  phoenix wavelength [micron]. Phenix native resolution
      ph_sed 			array
                  phenix spectral energy density [W m**-2 micron**-1]. Phenix native resolution
      ph_filename 		phenix filename
    """

    # def __init__(self, star_sed_path, star_distance, star_temperature, star_logg, star_f_h, star_radius):
    def __init__(self,
                 star_sed_path,
                 starDistance,
                 starTemperature,
                 starLogg,
                 starMetallicity,
                 starRadius,
                 use_planck_spectrum=False,
                 wl_min=0.4 * u.um,
                 wl_max=10.0 * u.um,
                 phoenix_model_filename=None):
        """
        Parameters
        __________
          exocat_star 	: 	object
                    exodata star object
          star_sed_path:    : 	string
                    path to Phoenix stellar spectra

        """
        self.set_log_name()

        if use_planck_spectrum == True:
            self.debug('Planck spectrum used')
            wl = np.linspace(wl_min, wl_max, 10000)
            ph_wl, ph_sed, ph_L = self.__get_star_spectrum(
                wl,
                starDistance.to(u.m),
                starTemperature.to(u.K),
                starRadius.to(u.m))
            ph_file = None
        else:
            if phoenix_model_filename:
                ph_file = os.path.join(star_sed_path, phoenix_model_filename)
                self.debug('phoenix file name : {}'.format(ph_file))
            else:
                ph_file = self.__get_phonix_model_filename(
                    star_sed_path,
                    starTemperature.to(u.K),
                    starLogg,
                    starMetallicity)
                self.debug('phoenix file name : {}'.format(ph_file))

            ph_wl, ph_sed, ph_L = self.__read_phenix_spectrum(
                ph_file,
                starDistance.to(u.m),
                starRadius.to(u.m))

        self.luminosity = ph_L
        self.sed = Sed(wl_grid=ph_wl, data=ph_sed)
        self.filename = ph_file
        self.model = 'Planck' if use_planck_spectrum else os.path.basename(ph_file)

    def __get_phonix_model_filename(self, path, star_temperature,
                                    star_logg, star_f_h):
        sed_name = glob.glob(os.path.join(path, "*.BT-Settl.spec.fits.gz"))

        if len(sed_name) == 0:
            self.error("No stellar SED files found")
            raise OSError("No stellar SED files found")

        sed_T_list = np.array([np.float(os.path.basename(k)[3:8]) for k in sed_name])
        sed_Logg_list = np.array([np.float(os.path.basename(k)[9:12]) for k in sed_name])
        sed_Z_list = np.array([np.float(os.path.basename(k)[13:16]) for k in sed_name])

        idx = np.argmin(np.abs(sed_T_list - np.round(star_temperature.value / 100.0)) +
                        np.abs(sed_Logg_list - star_logg) +
                        np.abs(sed_Z_list - star_f_h))

        ph_file = sed_name[idx]

        return ph_file

    def __read_phenix_spectrum(self, ph_file, star_distance, star_radius):

        """Read a PHENIX Stellar Spectrum.

        Parameters
        __________
          ph_file : 		string
                    full path to models file containing the SED
        Returns
        -------
          wl:			array
                    The Wavelength at which the SED is sampled. Units are [micron]
          sed :			array
                    The SED of the star. Units are [W m**-2 micron**-1]
          L :			scalar
                    The bolometric luminosity of the star. Units are [W]

        """

        ####################### USING PHOENIX BIN_SPECTRA BINARY FILES (h5)

        with fits.open(ph_file) as hdu:
            strUnit = hdu[1].header['TUNIT1']
            wl = hdu[1].data.field('Wavelength') * u.Unit(strUnit)

            strUnit = hdu[1].header['TUNIT2']
            sed = hdu[1].data.field('Flux') * u.Unit(strUnit)

            # remove duplicates
            idx = np.nonzero(np.diff(wl))
            wl = wl[idx]
            sed = sed[idx]

            bolometric_luminosity = (hdu[1].header['PHXLUM'] * u.W).to(u.Lsun)

            hdu.close()

        # Normalise SED to observed SED
        bolometric_flux = np.trapz(sed, x=wl)  # [W m**-2]
        bolometric_luminosity = 4 * np.pi * star_radius ** 2 * bolometric_flux  # [W]
        sed *= (star_radius / star_distance) ** 2  # [W/m^2/mu]

        # When trading two lines below for two above you'll be using the phoenix
        # luminosity, but this invalidates current validation tables against exosim
        # norm = bolometric_luminosity.to(u.W) / (4.0*np.pi*star_distance**2)
        # sed = norm * sed / bolometric_flux

        return wl, sed, bolometric_luminosity.to(u.Lsun)

    def __get_star_spectrum(self, wl, star_distance, star_temperature, star_radius):
        omega_star = np.pi * (star_radius / star_distance) ** 2 * u.sr
        sed = omega_star * exolib.planck(wl, star_temperature)
        bolometric_flux = np.trapz(sed, x=wl)
        bolometric_luminosity = 4.0 * np.pi * star_distance ** 2 * bolometric_flux
        return wl, sed, bolometric_luminosity.to(u.Lsun)


class CustomSed(Logger, object):

    def __init__(self, fname, star_radius, star_distance):
        self.set_log_name()
        ph = ascii.read(fname, format='ecsv')
        ph_wl = ph['Wavelength'].data * ph['Wavelength'].unit
        ph_sed = ph['Sed'].data * ph['Sed'].unit

        bolometric_flux = np.trapz(ph_sed, x=ph_wl)  # [W m**-2]
        bolometric_luminosity = 4 * np.pi * (star_radius.to(u.m)) ** 2 * bolometric_flux  # [W]
        ph_sed *= (star_radius.to(u.m) / star_distance.to(u.m)) ** 2  # [W/m^2/mu]
        self.debug('custom sed used : {}'.format(fname))

        self.sed = Sed(ph_wl, ph_sed)
        self.filename = fname
        self.luminosity = bolometric_luminosity.to(u.Lsun)
        self.model = 'Custom'



import logging

__all__ = ['Logger']

root_logger = logging.getLogger('exorad')
root_logger.setLevel(logging.DEBUG)

root_logger.propagate = False

class ExoRadHandler(logging.StreamHandler):

    def __init__(self, stream=None):
        from exorad.utils.mpi import get_rank
        super().__init__(stream=stream)

        self._rank = get_rank()

    def emit(self, record):
        # print(record)
        if self._rank == 0 or record.levelno >= logging.ERROR:
            # msg = '[{}] {}'.format(self._rank,record.msg)
            # record.msg = msg
            return super(ExoRadHandler, self).emit(record)
        else:
            pass


formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
ch = ExoRadHandler()
ch.setFormatter(formatter)
ch.setLevel(logging.INFO)
root_logger.addHandler(ch)


class Logger:
    """
    Standard logging using logger library
    """

    def __init__(self):
        self.set_log_name()

    def set_log_name(self):
        self._log_name = 'exorad.{}'.format(self.__class__.__name__)
        self._logger = logging.getLogger('exorad.{}'.format(self.__class__.__name__))

    def info(self, message, *args, **kwargs):
        """ See :class:`logging.Logger` """
        self._logger.info(message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        """ See :class:`logging.Logger` """
        self._logger.warning(message, *args, **kwargs)

    def debug(self, message, *args, **kwargs):
        """ See :class:`logging.Logger` """
        self._logger.debug(message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        """ See :class:`logging.Logger` """
        self._logger.error(message, *args, **kwargs)

    def critical(self, message, *args, **kwargs):
        """ See :class:`logging.Logger` """
        self._logger.critical(message, *args, **kwargs)



import astropy.units as u

from exorad.log import Logger
from exorad.models.signal import Radiance
from exorad.utils.exolib import planck


class ZodiacalFrg(Logger):
    """
    Produce the zodiacal radiance.

    Parameters
    ----------
    wl: array
        wavelength grid
    description: dict
        zodiacal foreground description
    coordinates: (float, float)
        pointing coordinates as (ra, dec)

    Attributes
    -----------
    wl: array
        wavelength grid
    description: dict
        zodiacal foreground description dict
    radiance: Radiance
        zodiacal radiance

    Methods
    --------
    model(A)
        returns the zodiacal Radiance for the zodiacal light model presented in Glasse et al. 2010,
        scaled by coefficient fitted over Kelsall et al. 1998 model
    zodiacal_fit_direction(coord)
        returns the fitted coefficient for the zodiacal model given the target position.
        It's based on Kelsall et al. 1998 model
    """

    def __init__(self, wl, description, coordinates=None):
        super().__init__()
        self.wl = wl
        self.description = description
        self.radiance = self._get_radiance(coordinates)

    def _get_radiance(self, coordinates):
        if coordinates and 'zodiacalMap' in self.description:
            if 'zodiacalMap':
                try:
                    A = self.zodiacal_fit_direction(coordinates)
                except (KeyError, OSError):
                    self.warning('zodiacal fit failed')
                    self._get_radiance(coordinates=None)

        elif 'zodiacFactor' in self.description:
            self.debug(' model used for zodiacal foreground')
            A = self.description['zodiacFactor']['value']
            self.debug('zodiac factor : {}'.format(A))
        else:
            self.warning('zodiacal description uncompleted')
            A = 0

        return self.model(A)

    def model(self, A):
        units = u.W / (u.m ** 2 * u.um * u.sr)
        zodi_emission = A * (3.5e-14 * planck(self.wl, 5500.0 * u.K) +
                             3.58e-8 * planck(self.wl, 270.0 * u.K)).to(units)
        return Radiance(self.wl, zodi_emission)

    def zodiacal_fit_direction(self, coord):
        import os
        from pathlib import Path
        from astropy.io.misc.hdf5 import read_table_hdf5
        import numpy as np

        ra_input = coord[0]
        dec_input = coord[1]

        dir_path = Path(os.path.dirname(os.path.realpath(__file__)))
        i = 0
        while 'data' not in [d.stem for d in Path(dir_path).iterdir() if d.is_dir()] or i > 10:
            dir_path = dir_path.parent
            i += 1
        if i > 10:
            self.error('Zodi map file not found')
            raise OSError('Zodi map file not found')

        data_path = os.path.join(dir_path.absolute().as_posix(), 'data')
        zodi_map_file = os.path.join(data_path, 'Zodi_map.hdf5')
        self.debug('map data:{}'.format(zodi_map_file))

        try:
            zodi_table = read_table_hdf5(zodi_map_file)
            self.debug(zodi_table)
            distance = (zodi_table['ra_icrs'] * u.deg - ra_input) ** 2 + (
                    zodi_table['dec_icrs'] * u.deg - dec_input) ** 2
            idx = np.argmin(distance)
            self.debug('selected line {}'.format(idx))
            self.debug(zodi_table[idx])
            return zodi_table['zodi_coeff'][idx]

        except OSError:
            self.error('Zodi map file not found')
            raise OSError('Zodi map file not found')



import datetime

import h5py
import hdfdict
from astropy.table import meta

from exorad.__version__ import __version__
from exorad.output.hdf5.util import recursively_read_dict_contents
from exorad.output.output import Output, OutputGroup

META_KEY = '__table_column_meta__'


class HDF5OutputGroup(OutputGroup):

    def __init__(self, entry):
        self.set_log_name()
        self._entry = entry

    def write_array(self, array_name, array, metadata=None):
        if isinstance(array, list):
            for idx, a in enumerate(array):
                self.write_array('{}{}'.format(array_name, idx), a, metadata)
            return
        ds = self._entry.create_dataset(str(array_name), data=array, shape=array.shape, dtype=array.dtype)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def write_table(self, table_name, table, metadata=None):
        table = _encode_mixins(table)
        if any(col.info.dtype.kind == 'U' for col in table.itercols()):
            table = table.copy(copy_data=False)
            table.convert_unicode_to_bytestring()

        self._entry.create_dataset(str(table_name), data=table.as_array())
        header_yaml = meta.get_yaml_from_table(table)

        header_encoded = [h.encode('utf-8') for h in header_yaml]
        self._entry.create_dataset(str(table_name) + '.' + META_KEY,
                                   data=header_encoded)

        tg = self._entry.create_group('{}_to_group'.format(str(table_name)))
        for col in table.keys():
            tg_c = tg.create_group(str(col))
            tg_c.create_dataset('value', data=table[col])
            if (table[col].unit != None):
                tg_c.create_dataset('unit', data=str(table[col].unit))

    def write_scalar(self, scalar_name, scalar, metadata=None):
        ds = self._entry.create_dataset(str(scalar_name), data=scalar)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def write_string(self, string_name, string, metadata=None):
        ds = self._entry.create_dataset(str(string_name), data=string)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def create_group(self, group_name):
        entry = None
        if self._entry:
            try:
                entry = self._entry.create_group(str(group_name))
            except ValueError:
                entry = self._entry[str(group_name)]
        return HDF5OutputGroup(entry)

    def write_string_array(self, string_name, string_array, metadata=None):

        asciiList = [n.encode("ascii", "ignore") for n in string_array]
        ds = self._entry.create_dataset(
            str(string_name), (len(asciiList), 1), 'S64', asciiList)

        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def write_quantity(self, quantity_name, quantity):
        if quantity_name == 'value':
            self._entry.create_dataset('value', data=quantity.value,)
        else:
            qg_c = self._entry.create_group(str(quantity_name))
            qg_c.create_dataset('value', data=quantity.value,)
            qg_c.create_dataset('unit', data=str(quantity.unit))
            pass

def _encode_mixins(tbl):
    from astropy.table import serialize
    from astropy.table.table import has_info_class
    from astropy import units as u
    from astropy.utils.data_info import MixinInfo, serialize_context_as

    try:
        import yaml
    except ImportError:
        for col in tbl.itercols():
            if (has_info_class(col, MixinInfo) and
                    col.__class__ is not u.Quantity):
                raise TypeError("cannot write type {} column '{}' "
                                "to HDF5 without PyYAML installed."
                                .format(col.__class__.__name__, col.info.name))

    with serialize_context_as('hdf5'):
        encode_tbl = serialize.represent_mixins_as_columns(tbl)

    return encode_tbl


class HDF5Output(Output):
    def __init__(self, filename, append=False):
        super().__init__()
        self.filename = filename
        self._append = append
        self.fd = None

    def open(self):
        self.fd = self._openFile(self.filename)

    def _openFile(self, fname):

        mode = 'w'
        if self._append:
            mode = 'a'

        fd = h5py.File(fname, mode=mode)
        fd.attrs['file_name'] = fname
        fd.attrs['file_time'] = datetime.datetime.now().isoformat()
        fd.attrs['creator'] = self.__class__.__name__
        fd.attrs['HDF5_Version'] = h5py.version.hdf5_version
        fd.attrs['h5py_version'] = h5py.version.version
        fd.attrs['program_name'] = 'ExoRad'
        fd.attrs['program_version'] = __version__
        return fd

    def create_group(self, group_name):
        entry = None
        if self.fd:
            try:
                entry = self.fd.create_group(str(group_name))
            except ValueError:
                entry = self.fd[str(group_name)]
        return HDF5OutputGroup(entry)

    def close(self):
        if self.fd:
            self.fd.flush()
            self.fd.close()


def load(input_group):
    """
    Reads and convert an HDF5 group into a dictionary

    Parameters
    ----------
    input_group: HDF5Group
        input dumped group

    Returns
    -------
    dict
    """
    input_group = hdfdict.load(input_group, lazy=False)
    input_group = recursively_read_dict_contents(input_group)
    return input_group



import os
from copy import deepcopy
from exorad.__version__ import __version__
from exorad.cache import GlobalCache
from exorad.log import disableLogging, enableLogging
from ..models.targetlist import XLXSTargetList, CSVTargetList, QTableTargetList
from .task import Task


class LoadTargetList(Task):
    """
    Loads target list from file

    Parameters
    ----------
    target_list: str
        target list file address. Supported formats are csv and xml


    Returns
    -------
    TargetList:
        return a list of Target class

    Examples
    --------
    >>> loadTargetList = LoadTargetList()
    >>> targets = loadTargetList(target_list='target_list/address')
    """

    def __init__(self):
        self.addTaskParam('target_list', 'target list file name')

    def execute(self):
        target_list_file = self.get_task_param('target_list')
        self.info('target list file : {}'.format(target_list_file))

        target_list_format = {".xlsx": XLXSTargetList,
                              ".csv": CSVTargetList,
                              }

        try:
            ext = os.path.splitext(target_list_file)[1]
            self.debug('target list format : {}'.format(ext))
            target_klass = target_list_format[ext]
        except KeyError:
            self.error('unsupported target list format: {}'.format(ext))
            raise IOError('unsupported target list format: {}'.format(ext))
        except TypeError:
            self.debug('target list is not a file. It is assumed to be QTable')
            target_klass = QTableTargetList

        tt = target_klass(target_list_file)
        self.set_output(tt)


class PrepareTarget(Task):
    """
    Prepares the target output table over the channels to populate with light propagation

    Parameters
    ----------
    target: Target
        target to prepare
    channels : dict
        channel dictionary

    Returns
    -------
    Target:
        same target with table attribute populated
    """

    def __init__(self):
        self.addTaskParam('target', 'target to prepare')
        self.addTaskParam('channels', 'channels dictionary')

    def execute(self):
        from . import MergeChannelsOutput

        target = self.get_task_param('target')
        channels = self.get_task_param('channels')

        mergeChannelsOutput = MergeChannelsOutput()
        table = mergeChannelsOutput(channels=channels)
        table = self.add_metadata(table, target)
        target.table = table
        self.set_output(target)

    def add_metadata(self, table, target):
        table.meta['name'] = target.name
        table.meta['ExoRad version'] = __version__
        return table


class UpdateTargetTable(Task):
    """
    Updated the target output table over the channels to populate with light propagation

    Parameters
    ----------
    target: Target
        target to prepare
    table : QTable
        table to merge in the target table

    Returns
    -------
    Target:
        same target with table attribute updated
    """

    def __init__(self):
        self.addTaskParam('target', 'target to prepare')
        self.addTaskParam('table', 'table to merge in the target table')

    def execute(self):
        from astropy.table import hstack
        self.info('updating target table')
        target = self.get_task_param('target')
        target_table = target.table
        table = self.get_task_param('table')

        repeated_keys = [key for key in target_table.keys() if key in table.keys()]
        if repeated_keys:
            self.debug('the following keys are already in both table and will be replaced : {}'.format(repeated_keys))
            target.table.remove_columns(repeated_keys)
        target.table = hstack([target_table, table])
        target.table.meta.update(table.meta)
        self.set_output(target)


class EstimateMaxSignal(Task):
    """
    Updated the target output table with maximum estimated signal in pix column

    Parameters
    ----------
    target: Target
        target to prepare

    Returns
    -------
    Target:
        same target with table attribute updated
    """

    def __init__(self):
        self.addTaskParam('target', 'target to prepare')

    def execute(self):
        import numpy as np
        import astropy.units as u
        from astropy.table import QTable
        updateTargetTable = UpdateTargetTable()

        self.info('estimating max signal in pixel')
        target = self.get_task_param('target')
        table = target.table
        signals = [key for key in table.keys() if 'Max' in key]
        max_signal = np.zeros(table['Wavelength'].size) * u.count / u.s
        for key in signals:
            max_signal += table[key]
        new_tab = QTable()
        new_tab['MaxSignal_inPixel'] = max_signal
        new_target = updateTargetTable(target=target, table=new_tab)
        self.set_output(new_target)


class ObserveTarget(Task):
    """
    Standard pipeline for target observation. It includes:
    1. PrepareTarget,
    2. EstimateForegrounds,
    3. PropagateForegroundLight
    4. LoadSource
    5. PropagateTargetLight
    6. EstimateNoise

    Parameters
    ----------
    target: Target
        target to prepare
    payload : dict
        payload description
    channels : dict
        channel dictionary
    wl_range: (float, float)
        wavelength range to investigate. (wl_min, wl_max)

    Returns
    -------
    Target:
        same target with table attribute updated
    """

    def __init__(self):
        self.addTaskParam('target', 'target to prepare')
        self.addTaskParam('payload', 'payload description')
        self.addTaskParam('channels', 'channel dictionary')
        self.addTaskParam('wl_range', 'wavelength range to investigate. (wl_min, wl_max)')

    def execute(self):
        target = self.get_task_param('target')
        payload = self.get_task_param('payload')
        channels = self.get_task_param('channels')
        wl_min, wl_max = self.get_task_param('wl_range')

        from . import PrepareTarget, PropagateTargetLight, \
            EstimateForegrounds, PropagateForegroundLight, EstimateNoise, LoadSource
        prepareTarget = PrepareTarget()
        propagateTargetLight = PropagateTargetLight()
        loadSource = LoadSource()
        estimateForegrounds = EstimateForegrounds()
        propagateForegroundLight = PropagateForegroundLight()
        estimateNoise = EstimateNoise()

        target = prepareTarget(target=target, channels=channels)

        if 'foreground' in payload['common']:
            target = estimateForegrounds(foregrounds=payload['common']['foreground'],
                                         target=target,
                                         wl_range=(wl_min, wl_max))
        target = propagateForegroundLight(channels=channels, target=target)

        target, sed = loadSource(target=target,
                                 source=payload['common']['sourceSpectrum'],
                                 wl_range=(wl_min, wl_max))
        target = propagateTargetLight(channels=channels, target=target)

        target = estimateNoise(target=target, channels=channels)

        self.set_output(target)


class ObserveTargetlist(Task):
    """
    Standard pipeline to observe a full targetlist. It allows parallelization:

    Parameters
    ----------
    targets: Target
        targets to prepare
    payload : dict
        payload description
    channels : dict
        channel dictionary
    wl_range: (float, float)
        wavelength range to investigate. (wl_min, wl_max)
    plot: bool
        allow to save plots
    out_dir: str
        indicate the directory where to save plots

    Returns
    -------
    dict:
        targets dict
    """

    def __init__(self):
        self.addTaskParam('targets', 'targets to prepare')
        self.addTaskParam('payload', 'payload description')
        self.addTaskParam('channels', 'channel dictionary')
        self.addTaskParam('wl_range', 'wavelength range to investigate. (wl_min, wl_max)')
        self.addTaskParam('plot', 'allow to save plots')
        self.addTaskParam('out_dir', 'indicate the directory where to save plots')

    def execute(self):
        from exorad.utils.util import chunks
        import multiprocessing as mp
        try:
            mp.set_start_method('fork')
        except RuntimeError:
            pass

        targets = self.get_task_param('targets')

        manager = mp.Manager()
        outputDict = manager.dict()
        for tt in chunks(targets, GlobalCache()['n_thread']):
            job = [mp.Process(target=self.pipeline_to_dict, args=(target, outputDict)) for target in tt]
            for j in job:
                j.start()
            for j in job:
                j.join()
        self.set_output(outputDict)

    def pipeline_to_dict(self,  target, outputDict):
        from . import ObserveTarget
        from exorad.utils.plotter import Plotter
        import matplotlib.pyplot as plt
        import matplotlib
        matplotlib.use('Agg')

        payload = self.get_task_param('payload')
        channels = self.get_task_param('channels')
        wl_range = self.get_task_param('wl_range')
        plot = self.get_task_param('plot')
        out_dir = self.get_task_param('out_dir')
        observeTarget = ObserveTarget()

        self.info('observing {}'.format(target.name))
        if not GlobalCache()['debug']: disableLogging()
        try:
            target = observeTarget(target=target, payload=payload, channels=channels, wl_range=wl_range)
            enableLogging()
            outputDict[target.name] = deepcopy(target)

            if plot:
                plotter = Plotter(input_table=target.table)
                plotter.plot_table()
                plotter.save_fig(os.path.join(out_dir, '{}.png'.format(target.name)))
                plt.close()

        except:
            enableLogging()
            self.warning('target {} skipped. Please check for previous error messages'.format(target.name))



from abc import ABC

import numpy as np

from exorad.log import Logger


class Output(Logger):

    def __init__(self):
        super().__init__()

    def open(self):
        raise NotImplementedError

    def create_group(self, group_name):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def store_dictionary(self, dictionary, group_name=None):
        from exorad.output.hdf5.util import recursively_save_dict_contents_to_output

        out = self
        if group_name is not None:
            out = self.create_group(group_name)

        recursively_save_dict_contents_to_output(out, dictionary)


class OutputGroup(Output, ABC):

    def __init__(self, name):
        super().__init__(name)
        self._name = name

    def write_array(self, array_name, array, metadata=None):
        raise NotImplementedError

    def write_list(self, list_name, list_array, metadata=None):
        arr = np.array(list_array)
        self.write_array(list_name, arr)

    def write_scalar(self, scalar_name, scalar, metadata=None):
        raise NotImplementedError

    def write_string(self, string_name, string, metadata=None):
        raise NotImplementedError

    def write_string_array(self, string_name, string_array, metadata=None):
        raise NotImplementedError



import copy
from abc import abstractmethod

import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.table import QTable
from scipy.interpolate import interp1d

from exorad.log.logger import Logger
from exorad.models.optics.opticalPath import OpticalPath
from exorad.models.signal import Signal
from exorad.utils.diffuse_light_propagation import prepare, convolve_with_slit, integrate_light
from exorad.models.utils import get_wl_col_name


class Instrument(Logger):
    """
    Abstract Instrument class. Contains the basic function that every instrument needs.

    Parameters
    -----------
    name: str
        instrument name
    description: dict
        instrument description dictionary
    payload: dict
        main payload. Default is None

    Attributes
    ----------
    name: str
        instrument name
    description: dict
        instrument description dictionary
    payload: dict
        main payload. Default is None
    table: QTable
        contain the output grid for the instrument
    built_instr: dict
        contains the instrument parameters needed to propagate the signal

    Methods
    --------
    built()
        it prepare the instrument to propagate the signal
    build_optical_path()
        it builds the instrument optical path
    propagate_target()
        propagates the target light trough the instrument
    propagate_diffuse_foreground()
        propagates the foreground light trough the instrument, starting for zodiacal light
    load(dir)
        it loads the instrument parameters already processed from a file
    write(dir)
        it writes the instrument parameters already processed from a file

    Raises
    -------
    ValueError:
        if you try to build a loaded payload.
    """

    def __init__(self, name, description, payload=None):
        self.set_log_name()
        self.name = name
        self.description = description
        self.payload = payload
        self.table = QTable()
        self.built_instr = dict()
        self.debug('{} initialized'.format(self.name))
        self.loaded = False
        self.opticalPath = None

    def load(self, table, built_instr):
        self.table = table
        self.built_instr = built_instr
        self.loaded = True
        self.info('loaded')

    def write(self, output):
        inst_out = output.create_group(self.name)
        inst_out.write_table(self.name, self.table, )
        inst_out.store_dictionary(self.description, group_name='description')
        inst_out.store_dictionary(self.built_instr, group_name='built_instr')
        self.info('instrument saved')
        return output

    def build(self):
        """
        check if it can build the instrument and populate the output table and dict.
        Then run the builder
        """

        if self.loaded:
            self.error('You cannot build a loaded instrument')
            # to build the instrument we need to be careful about the elements order, especially for optic emission!
            # you cannot load the description and run it again because it's not an ordered dictionary anymore
            raise ValueError('You cannot build a loaded instrument')
        else:
            self.builder()
            self.build_optical_path()

    def build_optical_path(self):

        self.info('building optical path')
        # what wl do I wanna use here?
        wl_grid = np.logspace(np.log10(self.description['detector']['wl_min']['value'].value),
                              np.log10(self.description['detector']['cut_off']['value'].value), 6000) * u.um

        common_optical_path = OpticalPath(wl=wl_grid, description=self.payload)
        channel_optical_path = OpticalPath(wl=wl_grid, description=self.description)
        channel_optical_path.prepend_optical_elements(common_optical_path.optical_element_dict)
        self.opticalPath = channel_optical_path

        transmission_table = self.opticalPath.build_transmission_table()
        optical_path_dict = {'transmission_table': transmission_table}
        # We need to rebin TR to the final one
        self.table['TR'], transmission_data = self._get_transmission(wl_grid, transmission_table['total'])
        self._add_data_to_built('transmission_data', transmission_data.to_dict())

        self.opticalPath.chain()
        if self.opticalPath.slit_width:
            self._add_data_to_built('slit_width', self.opticalPath.slit_width.to(u.um))

        self.table = self.opticalPath.compute_signal(self.table, self.built_instr)
        optical_path_dict['max_signal_per_pixel'] = self.opticalPath.max_signal_per_pixel
        optical_path_dict['signal_table'] = self.opticalPath.signal_table
        optical_path_dict['radiance_table'] = self.opticalPath.radiance_table
        self._add_data_to_built('optical_path', optical_path_dict)

    @abstractmethod
    def builder(self):
        """
        build the instrument and populate the output table and dict
        """
        pass

    @abstractmethod
    def propagate_target(self, target):
        """
        propagate point source
        """
        pass

    def propagate_diffuse_foreground(self, target):
        """
        propagate diffuse foreground sources, starting from zodiacal background
        """
        import copy
        self.debug('diffuse bkg propagation')
        out = QTable()
        total_max_signal, total_signal, wl_table, A, qe, omega_pix, transmission = prepare(self.table,
                                                                                           self.built_instr,
                                                                                           self.description)
        foregrounds = list(target.foreground.keys())
        foregrounds = reversed(foregrounds)
        for i, frg in enumerate(foregrounds):
            self.debug('propagating {}'.format(frg))
            radiance = copy.deepcopy(target.foreground[frg])
            self.debug('{} radiance . {}'.format(frg, radiance.data))

            if hasattr(frg, 'transmission'):
                frg.transmission.spectral_rebin(transmission.wl_grid)
                transmission.data *= frg.transmission.data
                self.debug('added {} transmission'.format(frg))

            transmission.spectral_rebin(radiance.wl_grid)
            radiance.data *= transmission.data

            # for other_el in foregrounds[i+1:]:
            #     if hasattr(target.foreground[other_el], 'transmission'):
            #         self.debug('passing through {}'.format(other_el))
            #         radiance.data *= target.foreground[other_el].transmission
            if 'slit_width' in self.built_instr:
                max_signal_per_pix, signal = convolve_with_slit(self.description, self.built_instr,
                                                                A, self.table, omega_pix, qe, radiance)
            else:
                qe.spectral_rebin(radiance.wl_grid)
                radiance.data *= omega_pix * A * qe.data * (qe.wl_grid / const.c / const.h).to(1. / u.W / u.s) * u.count
                # try:
                #     signal = np.trapz(radiance.data * self._window_function(radiance), x=radiance.wl_grid).to(u.ct / u.s)
                #     signal *= self.built_instr['window_spatial_width']
                # except KeyError:
                max_signal_per_pix, signal = integrate_light(radiance, radiance.wl_grid, self.built_instr)
            total_signal = copy.deepcopy(signal)
            self.debug('sed : {}'.format(total_signal))
            total_max_signal = copy.deepcopy(max_signal_per_pix)

            out['{}_signal'.format(frg)] = total_signal
            out['{}_MaxSignal_inPixel'.format(frg)] = total_max_signal
        return out

    def _bin_signal(self, wl, signal, leftbin, rightbin):

        bsig = [np.mean(signal[np.logical_and(wl >= wlow, wl < whigh)])
                for wlow, whigh in zip(leftbin, rightbin)]
        return u.Quantity(bsig)

    def _get_transmission(self, wl_grid, tr_data=None):

        tr_func = interp1d(wl_grid,
                           tr_data,
                           assume_sorted=False,
                           fill_value=0.0,
                           bounds_error=False)

        transmission_data = Signal(wl_grid, tr_data)
        # self.debug('transmission channel data : {} {}'.format(transmission_data.wl_grid, transmission_data.data))

        transmission = self._bin_signal(self.table['Wavelength'], tr_func(self.table['Wavelength']),
                                        self.table['LeftBinEdge'], self.table['RightBinEdge'])

        if self.payload['optics']['ForceChannelWlEdge']['value']:
            self.debug('force channel wl edge enabled')
            idx = np.logical_or(self.table['Wavelength']
                                < self.description['wl_min']['value'].to(self.table['Wavelength'].unit),
                                self.table['Wavelength']
                                > self.description['wl_max']['value'].to(self.table['Wavelength'].unit))
            transmission[idx] = 0.0
            idx = np.logical_or(transmission_data.wl_grid
                                < self.description['wl_min']['value'].to(self.table['Wavelength'].unit),
                                transmission_data.wl_grid
                                > self.description['wl_max']['value'].to(self.table['Wavelength'].unit))
            transmission_data.data[idx] = 0.0

        self.debug('transmission in channel : {}'.format(transmission))
        return transmission, transmission_data

    def _get_qe(self):
        if 'datafile' in self.description['detector']['qe']:
            self.debug('QE data file found')
            wl_colname = get_wl_col_name(self.description['detector']['qe']['data'])
            qe_data = Signal(self.description['detector']['qe']['data'][wl_colname], \
                             self.description['detector']['qe']['data'][self.name])
        else:
            wl_grid = np.logspace(np.log10(self.description['detector']['wl_min']['value'].to(u.um).value),
                                  np.log10(self.description['detector']['cut_off']['value'].to(u.um).value),
                                  6000) * u.um
            qe_data = Signal(wl_grid,
                             self.description['detector']['qe']['value'] * np.ones(wl_grid.size))

        qe_func = interp1d(qe_data.wl_grid,
                           qe_data.data,
                           assume_sorted=False,
                           fill_value=0.0,
                           bounds_error=False)
        qe = self._bin_signal(self.table['Wavelength'], qe_func(self.table['Wavelength']),
                              self.table['LeftBinEdge'], self.table['RightBinEdge'])
        self.debug('qe in channel : {}'.format(qe))
        return qe, qe_data

    def _get_efficiency(self, wl, target):
        # qe = interp1d(wl, self.built_instr['qe_data']['wl_grid'],
        #               self.built_instr['qe_data']['data'], left=0.0, right=0.0)
        # transmission = interp1d(wl, self.built_instr['transmission_data']['wl_grid'],
        #                         self.built_instr['transmission_data']['data'], left=0.0, right=0.0)
        # wave_window = np.ones(self.table['Wavelength'])
        qe = Signal(
            self.built_instr['qe_data']['wl_grid']['value'] * u.Unit(self.built_instr['qe_data']['wl_grid']['unit']),
            self.built_instr['qe_data']['data']['value'])
        qe.spectral_rebin(wl)

        transmission = Signal(self.built_instr['transmission_data']['wl_grid']['value']
                              * u.Unit(self.built_instr['transmission_data']['wl_grid']['unit']),
                              self.built_instr['transmission_data']['data']['value'])
        transmission.spectral_rebin(wl)

        if hasattr(target, 'skyTransmission'):
            target_transmission = copy.deepcopy(target.skyTransmission)
            self.table['sky TR'], _ = self._get_transmission(target_transmission.wl_grid, target_transmission.data)
            target_transmission.spectral_rebin(wl)
            transmission.data *= target_transmission.data

        wave_window = np.ones(wl.size)
        return qe.data, transmission.data, wave_window

    def _add_data_to_built(self, name, data):
        self.built_instr[name] = data

    def _window_function(self, signal):
        window_function = []
        for wld, wlu in zip(self.table['LeftBinEdge'], self.table['RightBinEdge']):
            mask = np.logical_and(signal.wl_grid >= wld,
                                  signal.wl_grid < wlu).astype(float)
            window_function.append(mask)
        window_function = np.array(window_function)
        self.debug('window function: {}'.format(window_function))
        return window_function



from .__about__ import (
    __author__,
    __commit__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __pkg_name__,
    __url__,
    __branch__,
    __base_dir__
)
from .__version__ import __version__

__all__ = [
    "__author__",
    "__commit__",
    "__copyright__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__pkg_name__",
    "__url__",
    "__version__",
    "__branch__",
    "__base_dir__"
]

from exorad.exorad import standard_pipeline

from exorad.cache import GlobalCache

gc = GlobalCache()
gc['n_thread'] = 1
gc['debug'] = False



import os.path
from datetime import date

__all__ = [
    '__pkg_name__',
    "__title__",
    "__summary__",
    "__url__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
    "__branch__",
    "__commit__",
    "__base_dir__"
]

__pkg_name__ = 'exorad'
__title__ = "ExoRad2"
__summary__ = "The generic point source radiometric model"
__url__ = "https://github.com/ExObsSim/ExoRad2-public"

try:
    base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
except NameError:
    base_dir = None

__base_dir__ = base_dir
__branch__ = None
if base_dir is not None and os.path.exists(os.path.join(base_dir, ".git")):
    git_folder = os.path.join(base_dir, ".git")
    with open(os.path.join(git_folder, "HEAD")) as fp:
        ref = fp.read().strip()
    ref_dir = ref[5:]
    __branch__ = ref[16:]
    try:
        with open(os.path.join(git_folder, ref_dir)) as fp:
            __commit__ = fp.read().strip()
    except FileNotFoundError:
        __commit__ = None
else:
    __commit__ = None


__author__ = "Lorenzo V. Mugnai, Enzo Pascale"
__email__ = "lorenzo.mugnai@uniroma1.it"

__license__ = "BSD-3-Clause"
__copyright__ = '2020-{:d}, {}'.format(date.today().year, __author__)



def get_wl_col_name(data, description=None):
    wl_colname = None
    if description is not None and 'wl_col_name' in description:
        wl_colname = description['wl_col_name']['value']
    else:
        for wl_key in ['Wavelength', 'wavelength']:
            if wl_key in data.keys():
                wl_colname= wl_key
    if wl_key is None:
        raise KeyError('Wavelength column not found in transmission data file')
    return wl_colname


import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.table import QTable
from scipy.interpolate import interp1d

from exorad.models.signal import CustomSignal, Signal, CountsPerSeconds
from exorad.utils.exolib import binnedPSF
from .instrument import Instrument


class Spectrometer(Instrument):
    """
    Spectrometer instrument class.
    """

    def _wavelength_table(self):
        # check if R is defined
        if 'targetR' not in self.description:
            self.warning('Channel targetR missing: native R is assumed')
            self.description['targetR'] = {'value':'native'}

        # define the wavelength grid
        if 'value' in self.description['targetR'].keys():
            self.debug('spectral resolution set to {}'.format(self.description['targetR']['value']))
            # native R
            if self.description['targetR']['value'] == 'native':
                wl_bin_c = self.built_instr['wl_pix_center']
                wl_bin_width = self.built_instr['pixel_bandwidth']
                wl_bin = wl_bin_c - 0.5 * wl_bin_width
                wl_bin = np.append(wl_bin, wl_bin_c[-1] + 0.5 * wl_bin_width[-1])
            # fixed R
            else:
                number_of_spectral_bins = np.ceil(
                    np.log(self.description['wl_max']['value']
                           / self.description['wl_min']['value'])
                    / np.log(1.0 + 1.0 / self.description['targetR']['value'])) + 1
                wl_bin = self.description['wl_min']['value'] \
                         * (1.0 + 1.0 / self.description['targetR']['value']) ** np.arange(number_of_spectral_bins)
                wl_bin_c = 0.5 * (wl_bin[0:-1] + wl_bin[1:])
                wl_bin_width = wl_bin[1:] - wl_bin[0:-1]
        # wavelength dependent R
        elif 'data' in self.description['targetR'].keys():
            raise NotImplementedError
        else:
            self.error('Channel targetR format unsupported.')
            raise KeyError('Channel targetR format unsupported.')

        # preparing the table
        self.table['chName'] = [self.name] * wl_bin_c.size
        self.table['Wavelength'] = wl_bin_c
        self.table['Bandwidth'] = wl_bin_width
        self.table['LeftBinEdge'] = self.table['Wavelength'] - 0.5 * self.table['Bandwidth']
        self.table['RightBinEdge'] = self.table['Wavelength'] + 0.5 * self.table['Bandwidth']
        self.debug('wavelength table: \n{}'.format(self.table))
        return wl_bin

    def builder(self):
        self.info('building {}'.format(self.name))

        # building pixel grid
        try:
            wl_solution_data = CustomSignal(
                self.description['wlSolution']['data']['Wavelength'],
                self.description['wlSolution']['data']['x'],
                self.description['wlSolution']['data']['x'].unit)
        except KeyError:
            self.error('Wavelength solution not indicated')
            raise

        self._add_data_to_built('wl_solution_data', wl_solution_data.to_dict())
        wl_sol_func = interp1d(wl_solution_data.data, wl_solution_data.wl_grid,
                               assume_sorted=False, fill_value='extrapolate', bounds_error=False)

        wl_sol_func_reverse = interp1d(wl_solution_data.wl_grid, wl_solution_data.data,
                                       assume_sorted=False, fill_value='extrapolate', bounds_error=False)

        first_pixel = wl_sol_func_reverse(self.description['wl_min']['value']) * wl_solution_data.data.unit
        last_pixel = wl_sol_func_reverse(self.description['wl_max']['value']) * wl_solution_data.data.unit
        self.debug('first pixel: {}, last pixel: {}'.format(first_pixel, last_pixel))
        if last_pixel < first_pixel: last_pixel, first_pixel = first_pixel, last_pixel

        delta = self.description['detector']['delta_pix']['value'].to(first_pixel.unit)

        # coordinates of detector  pixel centres
        coord_pix_center = np.arange(first_pixel.value, last_pixel.value, delta.value) * delta.unit
        coord_pix_center = np.flip(coord_pix_center)
        self.debug('pix_center: {}'.format(coord_pix_center))
        self._add_data_to_built('pix_center', coord_pix_center)
        # wavelength sampled by the pixels
        pixel_wavelength = wl_sol_func(coord_pix_center) * wl_solution_data.wl_grid.unit
        self.debug('wl_pix_center: {}'.format(pixel_wavelength))
        self._add_data_to_built('wl_pix_center', pixel_wavelength)
        pixel_bandwidth = np.abs(wl_sol_func(coord_pix_center - 0.5 * delta) -
                                 wl_sol_func(coord_pix_center + 0.5 * delta)) * delta.unit
        self._add_data_to_built('pixel_bandwidth', pixel_bandwidth)

        # initializing channel table
        wl_bin = self._wavelength_table()
        self._add_data_to_built('wl_bin', wl_bin)
        self.table['QE'], qe_data = self._get_qe()
        self._add_data_to_built('qe_data', qe_data.to_dict())

        # window sizes
        pixel_window_edge = wl_sol_func_reverse(wl_bin) * wl_solution_data.data.unit
        window_spectral_width = (np.abs(pixel_window_edge[1:] - pixel_window_edge[0:-1]) / delta).to('')

        if 'EncESolution' in self.description.keys():
            self.debug('Encircle energy solution found')
            EncE_sol = interp1d(self.description['EncESolution']['data']['Wavelength'],
                                self.description['EncESolution']['data']['EE'],
                                assume_sorted=False)
            radius = EncE_sol(self.table['Wavelength'])
        else:
            radius = 1.22
        self.debug('radius : {}'.format(radius))

        window_spatial_width = (2.0 * radius * self.description['Fnum_y']['value'] *
                                self.table['Wavelength'] / delta).to('')
        if 'window_spatial_scale' in list(self.description.keys()):
            window_spatial_width *= self.description['window_spatial_scale']['value']
            self.debug('window spatial width scaled by {}'.format(self.description['window_spatial_scale']['value']))

        window_size_px = window_spectral_width * window_spatial_width
        self._add_data_to_built('window_spectral_width', window_spectral_width)
        self._add_data_to_built('window_spatial_width', window_spatial_width)
        self._add_data_to_built('window_size_px', window_size_px)
        self.table['WindowSize'] = window_size_px
        self.debug('window size : {}'.format(window_size_px))

        # PSF gain
        _gain_prf = np.empty(10, dtype=float)
        wl_prf = np.linspace(pixel_wavelength.min(), pixel_wavelength.max(),
                             _gain_prf.size)
        for k, wl in enumerate(wl_prf):
            prf, pixel_prf, extent = binnedPSF(self.description['Fnum_x']['value'],
                                               self.description['Fnum_y']['value'],
                                               wl,
                                               delta)
            _gain_prf[k] = np.max(prf.sum(axis=0) / pixel_prf.sum(axis=0).max())

            if 'WFErms' in self.description.keys():
                self.debug('WFE found')
                Strehl = np.exp(- (2 * np.pi * self.description['WFErms']['value'] / wl) ** 2)
            else:
                Strehl = 1.0
            _gain_prf[k] *= Strehl
        self.debug(_gain_prf)
        gain_prf_data = Signal(wl_prf, _gain_prf)
        self._add_data_to_built('gain_prf_data', gain_prf_data.to_dict())

    def propagate_target(self, target):
        out = QTable()
        wl = target.star.sed.wl_grid
        qe, transmission, wave_window = self._get_efficiency(wl, target)
        if 'sky TR' in self.table.keys():
            out['foreground_transmission'] = self.table['sky TR']
        if self.payload['optics']['ForceChannelWlEdge']['value']:
            self.debug('force channel wl edge enabled')
            idx = np.logical_or(wl < self.description['wl_min']['value'].to(self.table['Wavelength'].unit),
                                wl > self.description['wl_max']['value'].to(self.table['Wavelength'].unit))
            transmission[idx] = wave_window[idx] = 0.0

        signal_density = CustomSignal(wl_grid=target.star.sed.wl_grid,
                                      data=(self.payload['optics']['Atel']['value']
                                            * transmission
                                            * qe
                                            * target.star.sed.data
                                            * target.star.sed.wl_grid.to(u.m) /
                                            const.h / const.c).to(1 / u.um / u.s) * u.count,
                                      data_unit=u.count / u.um / u.s)
        self.debug('star signal density: {}'.format(signal_density.data))

        # max signal in pixel
        gain_prf_data = self.built_instr['gain_prf_data']
        gain_prf = interp1d(gain_prf_data['wl_grid']['value'], gain_prf_data['data']['value'],
                            kind='cubic', assume_sorted=False)

        # signal in spectral bin
        window_function = self._window_function(target.star.sed)
        self.debug('window function: {}'.format(window_function))

        flux_density = wave_window * target.star.sed.data
        self.debug('star flux density: {}'.format(flux_density))

        star_flux = np.trapz(flux_density * window_function,
                             x=target.star.sed.wl_grid.to(u.um)).to(u.W / u.m ** 2)
        self.debug('star flux : {}'.format(star_flux))
        out['starFlux'] = star_flux

        star_signal = np.trapz(signal_density.data * window_function,
                               x=target.star.sed.wl_grid).to(u.count / u.s)
        self.debug('star signal : {}'.format(star_signal))
        out['starSignal'] = star_signal
        out['star_signal_inAperture'] = star_signal

        star_signal_inPixel_density = signal_density
        star_signal_inPixel_density.spectral_rebin(self.built_instr['wl_pix_center'])

        star_signal_inPixel = CountsPerSeconds(star_signal_inPixel_density.wl_grid,
                                               (star_signal_inPixel_density.data * self.built_instr['pixel_bandwidth']
                                                * gain_prf(star_signal_inPixel_density.wl_grid)).to(u.count / u.s))
        starSignal_inPixel_max = np.empty(self.table['Wavelength'].size, dtype=float) * star_signal_inPixel.data.unit
        for k, (wld, wlu) in enumerate(zip(self.table['LeftBinEdge'], self.table['RightBinEdge'])):
            idx = np.where(np.logical_and(star_signal_inPixel.wl_grid > wld,
                                          star_signal_inPixel.wl_grid < wlu))[0]
            starSignal_inPixel_max[k] = np.max(star_signal_inPixel.data[idx])
        out['star_MaxSignal_inPixel'] = starSignal_inPixel_max
        self.debug('star signal in pixel MAX : {}'.format(starSignal_inPixel_max))
        return out
